from django import forms
from edc_action_item.forms import ActionItemFormMixin
from edc_crf.crf_form_validator import CrfFormValidator
from edc_crf.modelform_mixins import CrfModelFormMixin
from edc_lab_panel.panels import fbc_panel, hba1c_panel

from edc_lab_results.form_validator_mixins import BloodResultsFormValidatorMixin

from .models import BloodResultsFbc, BloodResultsHba1c


class BloodResultsFbcFormValidator(BloodResultsFormValidatorMixin, CrfFormValidator):
    panel = fbc_panel

    def validate_demographics(self) -> None:
        """Skip for tests"""
        pass


class BloodResultsHba1cFormValidator(BloodResultsFormValidatorMixin, CrfFormValidator):
    panel = hba1c_panel

    def validate_demographics(self) -> None:
        """Skip for tests"""
        pass


class BloodResultsFbcForm(ActionItemFormMixin, CrfModelFormMixin, forms.ModelForm):
    form_validator_cls = BloodResultsFbcFormValidator

    def validate_against_consent(self):
        """Skip for tests"""
        pass

    class Meta:
        model = BloodResultsFbc
        fields = "__all__"


class BloodResultsHba1cForm(CrfModelFormMixin, forms.ModelForm):
    form_validator_cls = BloodResultsHba1cFormValidator

    def validate_against_consent(self):
        """Skip for tests"""
        pass

    class Meta:
        model = BloodResultsHba1c
        fields = "__all__"



from typing import Any

from edc_glucose.utils import validate_glucose_as_millimoles_per_liter

from .blood_results_fbg_form_validator_mixin import BloodResultsFbgFormValidatorMixin


class BloodResultsGluFormValidatorMixin(BloodResultsFbgFormValidatorMixin):
    def evaluate_value(self: Any, field_name: str):
        if field_name == "glucose_value":
            validate_glucose_as_millimoles_per_liter("glucose", self.cleaned_data)



from django.apps import apps as django_apps
from django.test import TestCase
from edc_appointment.models import Appointment
from edc_constants.constants import NO, NOT_APPLICABLE, YES
from edc_lab.models import Panel
from edc_reportable import GRADE3, GRAMS_PER_DECILITER, PERCENT, TEN_X_9_PER_LITER
from edc_utils import get_utcnow
from edc_visit_tracking.constants import SCHEDULED

from ..models import BloodResultsFbc
from ..test_case_mixin import TestCaseMixin


class TestBloodResult(TestCaseMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.subject_identifier = self.enroll()
        appointment = Appointment.objects.get(
            subject_identifier=self.subject_identifier,
            visit_code="1000",
        )
        subject_visit = django_apps.get_model(
            "edc_visit_tracking.subjectvisit"
        ).objects.create(
            report_datetime=get_utcnow(),
            appointment=appointment,
            reason=SCHEDULED,
        )
        panel = Panel.objects.get(name="fbc")
        requisition = django_apps.get_model("edc_metadata.subjectrequisition").objects.create(
            subject_visit=subject_visit,
            panel=panel,
            requisition_datetime=subject_visit.report_datetime,
        )
        self.data = dict(subject_visit=subject_visit, requisition=requisition)

    def test_ok(self):
        BloodResultsFbc.objects.create(**self.data)

    def test_summary_none(self):
        obj = BloodResultsFbc.objects.create(**self.data)
        self.assertEqual(obj.get_summary(), [])

    def test_summary_normal(self):
        self.data.update(
            haemoglobin_value=14,
            haemoglobin_units=GRAMS_PER_DECILITER,
            haemoglobin_abnormal=NO,
            haemoglobin_reportable=NOT_APPLICABLE,
            results_abnormal=NO,
            results_reportable=NOT_APPLICABLE,
        )
        obj = BloodResultsFbc.objects.create(**self.data)
        summary = obj.get_summary()
        self.assertEqual([], summary)

    def test_summary_abnormal(self):
        self.data.update(
            haemoglobin_value=12,
            haemoglobin_units=GRAMS_PER_DECILITER,
            haemoglobin_abnormal=YES,
            haemoglobin_reportable=GRADE3,
            results_abnormal=YES,
            results_reportable=NO,
        )
        obj = BloodResultsFbc.objects.create(**self.data)
        summary = obj.get_summary()
        self.assertIn("haemoglobin_value: 12 g/dL is abnormal", summary)

    def test_summary_g3(self):
        self.data.update(
            haemoglobin_value=7.5,
            haemoglobin_units=GRAMS_PER_DECILITER,
            haemoglobin_abnormal=YES,
            haemoglobin_reportable=GRADE3,
            results_abnormal=YES,
            results_reportable=YES,
        )
        obj = BloodResultsFbc.objects.create(**self.data)
        summary = obj.get_summary()
        self.assertIn("haemoglobin_value: 7.0<=7.5<9.0 g/dL GRADE 3.", summary)

    def test_missing(self):
        obj = BloodResultsFbc.objects.create(**self.data)
        self.assertEqual(obj.missing_count, 5)
        self.assertEqual(
            "haemoglobin_value,hct_value,rbc_value,wbc_value,platelets_value", obj.missing
        )

        obj.haemoglobin_value = 14
        obj.haemoglobin_units = GRAMS_PER_DECILITER
        obj.save()
        self.assertEqual(obj.missing_count, 4)
        self.assertEqual("hct_value,rbc_value,wbc_value,platelets_value", obj.missing)

        obj.hct_value = 10
        obj.hct_units = PERCENT
        obj.save()
        self.assertEqual(obj.missing_count, 3)
        self.assertEqual("rbc_value,wbc_value,platelets_value", obj.missing)

        obj.rbc_value = 10
        obj.rbc_units = TEN_X_9_PER_LITER
        obj.save()
        self.assertEqual(obj.missing_count, 2)
        self.assertEqual("wbc_value,platelets_value", obj.missing)

        obj.wbc_value = 10
        obj.wbc_units = TEN_X_9_PER_LITER
        obj.save()
        self.assertEqual(obj.missing_count, 1)
        self.assertEqual("platelets_value", obj.missing)

        obj.platelets_value = 10
        obj.platelets_units = TEN_X_9_PER_LITER
        obj.save()
        self.assertEqual(obj.missing_count, 0)
        self.assertEqual(None, obj.missing)

        obj.platelets_value = None
        obj.platelets_units = None
        obj.save()
        self.assertEqual(obj.missing_count, 1)
        self.assertEqual("platelets_value", obj.missing)



from dateutil.relativedelta import relativedelta
from django.test import TestCase
from edc_action_item import site_action_items
from edc_consent import site_consents
from edc_consent.tests.consent_test_utils import consent_object_factory
from edc_facility.import_holidays import import_holidays
from edc_lab import site_labs
from edc_metadata.tests.models import SubjectConsent
from edc_registration.models import RegisteredSubject
from edc_reportable import site_reportables
from edc_reportable.grading_data.daids_july_2017 import grading_data
from edc_reportable.normal_data.africa import normal_data
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from edc_lab_results.action_items import register_actions

from .lab_profiles import subject_lab_profile
from .visit_schedule import visit_schedule


class TestCaseMixin(TestCase):
    @classmethod
    def setUpClass(cls):
        site_reportables._registry = {}
        site_labs.initialize()
        site_action_items.registry = {}
        site_visit_schedules._registry = {}
        site_consents.registry = {}
        super().setUpClass()

    @classmethod
    def setUpTestData(cls):
        site_visit_schedules.register(visit_schedule)
        consent_object_factory(
            model=SubjectConsent._meta.label_lower,
            start=get_utcnow() - relativedelta(years=1),
            end=get_utcnow() + relativedelta(years=1),
        )
        import_holidays()
        site_reportables.register(
            name="my_reportables", normal_data=normal_data, grading_data=grading_data
        )
        site_labs.register(lab_profile=subject_lab_profile)
        register_actions()

    @staticmethod
    def enroll(subject_identifier=None):
        subject_identifier = subject_identifier or "1111111"
        subject_consent = SubjectConsent.objects.create(
            subject_identifier=subject_identifier, consent_datetime=get_utcnow()
        )
        _, schedule = site_visit_schedules.get_by_onschedule_model("edc_metadata.onschedule")
        schedule.put_on_schedule(
            subject_identifier=subject_consent.subject_identifier,
            onschedule_datetime=subject_consent.consent_datetime,
        )
        return subject_identifier

    @staticmethod
    def fake_enroll():
        subject_identifier = "2222222"
        RegisteredSubject.objects.create(subject_identifier=subject_identifier)
        return subject_identifier
